
#ifndef __RESULT_H
#define __RESULT_H

#include <boost/type_traits.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <errno.h>
#include <iostream>

#include "Atomic.h"

#define CACHE_LINE_SIZE 64

extern bool do_thread_task( void );

class AbstractResult {
public:
  virtual ~AbstractResult(){}
  virtual void run() = 0;
};

/**
 * Result class; ought to work for most C++ types,
 * but definitely will work best for simple data types.
 * If we need to be passing classes around a lot,
 * we might want to make some changes.
 */
template < typename T >
class real_result {//: public AbstractResult {

 private:
  T outcome;
  bool has_run, has_refed;
  pthread_mutex_t running_mutex;
  int num_dels;
  //  tbb::atomic<int> num_refs;
  int num_refs;

 public:

  /* used for the real_result freelist */
  static struct {
    pthread_key_t key;
    bool initialized;
    char pad[CACHE_LINE_SIZE - sizeof(pthread_key_t) - sizeof(bool)];
  } freelist __attribute__ ((aligned(CACHE_LINE_SIZE)));

  real_result* next;

  boost::function <T (void)> future; // future isn't modified after the call to task

  real_result(){
    has_run = has_refed = false;
    num_refs = num_dels = 0;
    pthread_mutex_init(&running_mutex, NULL);
    next = NULL;
  }

  ~real_result(){
#ifdef TASKMAN_DEBUG
    assert( num_dels == 0 );
    assert( has_run );
    assert( num_refs == 0 );
#endif

    num_dels++;
    pthread_mutex_destroy( &running_mutex );

  }

  real_result(const real_result & obj)
  {
    assert(false); // should never copy the real result
  }

  real_result<T>& operator=( const real_result<T>& t ){
    assert(false); // should never call
    return *this;
  }

  void addReference()
  {
    //    num_refs.fetch_and_add(1);    
    fetchAndAdd(num_refs, 1);
  }

  int removeReference()
  {
    //    int result = num_refs.fetch_and_add(-1);
    int result = fetchAndAdd(num_refs, -1);
    return result - 1;
  }

  T operator*(){ 
    if( ! has_run ){ 
      std::cerr << "operator* on a real_result that hasn't run!  (You're in trouble.)" << std::endl;
      run();
    }
    if( !has_refed ){
      removeReference();
      has_refed = true;
    }
    return outcome; 
  }
  
  void run(){
    
    while( !has_run ){
      if(pthread_mutex_trylock(&running_mutex) == 0 ){
        if(!has_run) {
          outcome = future();
          has_run = true;
        }
        pthread_mutex_unlock(&running_mutex);
      }
      
    }
  }

  bool hasRun() const { 
    return has_run;
  }
};

template < typename T >
class result : public AbstractResult {
private:
  real_result<T>* r;

public:
  static struct {
    pthread_key_t key;
    bool initialized;
    char pad[CACHE_LINE_SIZE - sizeof(pthread_key_t) - sizeof(bool)];
  } freelist __attribute__ ((aligned(CACHE_LINE_SIZE)));
  result* next;

  result() : r(NULL), next(NULL) { }
    result( const result<T> & obj) : AbstractResult(), r(obj.r) { r->addReference(); } 
    result( real_result<T>* _r) : AbstractResult(), r(_r) {  r->addReference(); }
  ~result() {
    if(r != NULL){
      int num_refs;
      num_refs = r->removeReference();
#ifdef TASKMAN_DEBUG
      if(num_refs <= 0) {
        if( num_refs != 0 ){ std::cerr << "Alert! num_refs below zero." << std::endl; }
        //std:: cerr << "Deleting " << r << std::endl;
        delete r;
      }
#endif
    }
  }

  result<T>& operator=( const result<T>& res ) {
      
      if( this != &res ){
#ifdef TASKMAN_DEBUG
          if( r != NULL ){
              std::cerr << "Operator= on an active result" << std::endl;
              exit(EXIT_FAILURE);
          }
          assert( r != res.r && res.r != NULL );
#endif          

          r = res.r;
          r->addReference();
          
      }
      return *this;
  }

  void setFuture( boost::function<T (void)>& future) {
#ifdef TASKMAN_DEBUG
    if(r != NULL)
#endif
      r->future = future;
#ifdef TASKMAN_DEBUG
    else{
      std::cerr << "Attempted to use result before assignment" << std::endl;
      exit(EXIT_FAILURE);
    }
#endif
  }

  void run() { 
#ifdef TASKMAN_DEBUG
    if(r != NULL)
#endif
      r->run(); 
#ifdef TASKMAN_DEBUG
    else{
      std::cerr << "Attempted to use result before assignment" << std::endl;
      exit(EXIT_FAILURE);
    }
#endif
  }
  T operator*(){ 
#ifdef TASKMAN_DEBUG
    if(r != NULL){
#endif      

      while( !r->hasRun() ){
        do_thread_task();
      }
      return *(*r);
#ifdef TASKMAN_DEBUG      
    }
    else{
      std::cerr << "Attempted to use result before assignment" << std::endl;
      exit(EXIT_FAILURE);
    }
#endif
  }
};

/** 
 * I briefly worked on a template specialization for 
 * result<void>, but I'm leaving it unfinished for now.
 */
#if 0
template <>
class result< void > : public AbstractResult {
 public:
  result(){}
  ~result(){}

};
#endif // 0 

#endif // __RESULT_H
