
#include <pthread.h>
#include "TaskMan.h"
#include "Atomic.h"
#include "XactionQueue.h"
#include "FineGrainedLockQueue.h"
#include "CoarseGrainedLockQueue.h"

#define TASKMAN_ALLOCATION_CHUNK_SIZE 1024

#if defined(__sparc) || defined(__sparc__) || defined(sparc)
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>

static const int chianti_binds[32] = {0,4,8,12,16,20,24,28,1,5,9,13,17,21,25,29,2,6,10,14,18,22,26,30,3,7,11,15,19,23,27,31 };
#endif

#ifdef TLS_SUPPORTED

__thread AbstractQueue<AbstractResult*>* g_task_queue;

#else

static pthread_key_t queue_key;
AbstractQueue< AbstractResult*>* get_g_task_queue();

#endif

bool g_queue_empty;
queue_type_t g_queue_type = DEFAULT_QUEUE_TYPE;
volatile int g_idle_threads = 0;
#if defined(__linux) || defined(__linux__) || defined(linux)
double  g_cpu_mhz;
#endif

int g_num_threads;

/* these are shared variables that rarely change.  they are bundled to reduce false sharing misses */
static struct {
  pthread_t* threads;
  bool stop;
  bool initialized;
  bool timers_initialized;
  char pad[CACHE_LINE_SIZE-sizeof(pthread_t*)-sizeof(bool)-sizeof(bool)-sizeof(bool)];
} taskman_flags __attribute__ ((aligned(CACHE_LINE_SIZE)));

/*
static pthread_t* taskman_threads;
static bool stop = false;

static bool initialized = false;
static bool timers_initialized = false;
*/
#ifndef TLS_SUPPORTED

static void initialize_queue_key(){

  pthread_key_create( &queue_key, NULL );

}

AbstractQueue <AbstractResult*>* get_g_task_queue(){

#ifdef TASKMAN_DEBUG
  assert( taskman_flags.initialized );
#endif

  AbstractQueue< AbstractResult* >* p;
  p = static_cast< AbstractQueue<AbstractResult*>* > (pthread_getspecific( queue_key )) ;
  return p;

}

/*
template < typename T >
inline real_result<T>* get_real_result_freelist(){

  if( ! real_result<T>::freelist.initalized ){
    real_result<T>::freelist.initalized = true;
    pthread_key_create( & real_result<T>::freelist.key, NULL);
    pthread_setspecific( real_result<T>::freelist.key, new real_result<T>[TASKMAN_ALLOCATION_SIZE] );
  }

  real_result<T>* res;
  res = static_cast< real_result* > ( pthread_getspecific( real_result<T>::freelist.key ) );
  return res;

}
*/

template < typename T >
inline void set_real_result_freelist( real_result<T>* res ){

  if( ! real_result<T>::freelist.initalized ){
    real_result<T>::freelist.initalized = true;
    pthread_key_create( & real_result<T>::freelist.key, NULL);
  }

  pthread_setspecific( real_result<T>::freelist.key, static_cast< void* > (res) );

}

/*
template < typename T >
inline result<T>* get_result_freelist(){

  if( ! result<T>::freelist.initalized ){
    result<T>::freelist.initalized = true;
    pthread_key_create( & result<T>::freelist.key, NULL);
    pthread_setspecific( result<T>::freelist.key, new result<T>[TASKMAN_ALLOCATION_SIZE] );
  }

  result<T>* res;
  res = static_cast< real_result* > ( pthread_getspecific( result<T>::freelist.key ) );
  return res;

}
*/

template < typename T >
inline void set_result_freelist( result<T>* res ){

  if( ! result<T>::freelist.initalized ){
    result<T>::freelist.initalized = true;
    pthread_key_create( & result<T>::freelist.key, NULL);
  }

  pthread_setspecific( result<T>::freelist.key, static_cast< void* > (res) );

}

#endif

void* task_loop(void* args);

AbstractQueue<AbstractResult*>* new_task_queue()
{
#ifdef TASKMAN_DEBUG
  assert( taskman_flags.initialized );
#endif

  AbstractQueue<AbstractResult*>* ret = NULL;
  switch(g_queue_type){
  case XACTION_QUEUE:
    ret = new XactionQueue<AbstractResult*>();
    break;
  case FINE_LOCK_QUEUE:
    ret = new FineGrainedLockQueue<AbstractResult*>();
    break;
  case COARSE_LOCK_QUEUE:
    ret = new CoarseLockQueue<AbstractResult*>();
    break;
  case LOCK_FREE_QUEUE:
    std::cerr << "LOCK_FREE_QUEUE not implemented! (Crashes and death ahead.)" << std::endl; 
    break; 
  case HARDWARE_QUEUE:
    std::cerr << "HARDWARE_QUEUE not implemented! (Explosions and misery ahead.)" << std::endl;
    break;
  }
  return ret;
}

template < typename T >
real_result<T>* new_real_result()
{
#ifdef TASKMAN_DEBUG
  assert( taskman_flags.initialized );
#endif

  real_result<T>* res;
  real_result<T>* freelist;

  if( ! real_result<T>::freelist.initalized ){
    real_result<T>::freelist.initalized = true;
    pthread_key_create( & real_result<T>::freelist.key, NULL);
    pthread_setspecific( real_result<T>::freelist.key, new real_result<T>[TASKMAN_ALLOCATION_CHUNK_SIZE] );
  }
  freelist = static_cast< real_result<T>* > ( pthread_getspecific( real_result<T>::freelist.key ) );

  if( freelist->next == NULL ){
    // the freelist only has one element left...grow 
    freelist->next = new real_result<T>[TASKMAN_ALLOCATION_CHUNK_SIZE];
  }

  res = freelist;
  set_real_result_freelist( res->next );
  return res;  
  
}

template < typename T >
result<T>* new_result()
{
#ifdef TASKMAN_DEBUG
  assert( taskman_flags.initialized );
#endif

  result<T>* res;
  result<T>* freelist;

  if( ! result<T>::freelist.initalized ){
    result<T>::freelist.initalized = true;
    pthread_key_create( & result<T>::freelist.key, NULL);
    pthread_setspecific( result<T>::freelist.key, new result<T>[TASKMAN_ALLOCATION_CHUNK_SIZE] );
  }

  freelist = static_cast< result<T>* > ( pthread_getspecific( result<T>::freelist.key ) );

  if( freelist->next == NULL ){
    // the freelist only has one element left...grow 
    freelist->next = new result<T>[TASKMAN_ALLOCATION_CHUNK_SIZE];

  }

  res = freelist;
  set_result_freelist( res->next );
  return res;  
  
}

void taskman_sync()
{
  while(do_thread_task()) {}
}

static pthread_mutex_t barrier_lock;
static volatile bool global_barrier_sense = false;
static volatile int global_barrier_count = 0;

void task_barrier( bool* local_sense )
{
#ifdef TASKMAN_DEBUG
  assert( taskman_flags.initialized );
#endif

  *local_sense = !(*local_sense);

  pthread_mutex_lock( &barrier_lock );

  global_barrier_count++;

  pthread_mutex_unlock( &barrier_lock );

  if( global_barrier_count == (g_num_threads+1) ){
    global_barrier_count = 0;
    global_barrier_sense = *local_sense;
  }
  else{
    while( *local_sense != global_barrier_sense ) /*spin*/ ;
  }
  

}

// do_thread_task declared in Result.h
bool do_thread_task( void ){

  AbstractResult* res;

  if( g_task_queue->dequeue( res ) ){
    res->run();
    delete res;
    return true;
  }
  else{
    return false;
  }

}

static bool do_thread_task( AbstractQueue< AbstractResult* > * queue ){

  AbstractResult* res;

  if( queue->dequeue( res ) ){
    res->run();
    delete res;
    return true;
  }
  else{
    return false;
  }

}

void* task_loop(void* args)
{
  //int thread_id = reinterpret_cast<int> (args);  
  
#if defined(__sparc) || defined(__sparc__) || defined(sparc)
  if ( processor_bind( P_LWPID, P_MYID, chianti_binds[reinterpret_cast<long long int> (args) + 1], NULL ) ){
    std::cerr << "Error from processor_bind.\n" << std::endl;
  }
  //else{
  //  std::cerr << "Bound processor " << chianti_binds[((int)args)+1] << std::endl;
  //}
#endif

  bool local_sense = false;
  AbstractQueue <AbstractResult*> * queue;

  task_barrier( &local_sense );
  
#ifdef TLS_SUPPORTED
  queue = g_task_queue = new_task_queue();
#else
  queue = new_task_queue();
  pthread_setspecific( queue_key, queue );
#endif

  while(!taskman_flags.stop){

    while( do_thread_task( queue ) ){ }

    //if(g_queue_empty != 1){
    //  g_queue_empty = 1; 
    //}
    //fetchAndAdd(g_idle_threads, 1);
    //while(g_queue_empty){
      // spin until an enqueue occurs
      // I purposely didn't lock access to this flag
      //   ...I don't think it matters
      //   ...but, I'd like a second opinion
    //}
    //fetchAndAdd(g_idle_threads, -1);
  }

  return 0;
}

void set_queue_type(queue_type_t type) { 
  assert( !taskman_flags.initialized );
  g_queue_type = type; 
}


const char* get_queue_type_name() {
  switch( g_queue_type ){
  case COARSE_LOCK_QUEUE:
    return "Coarse-grained locking"; 
  case FINE_LOCK_QUEUE:
    return "Fine-grained locking";
  case XACTION_QUEUE:
    return "Transactional memory";
  case LOCK_FREE_QUEUE:
    return "Lock-free data structures";
  case HARDWARE_QUEUE:
    return "'Carbon' hardware mechanism";
  }
  return "Something strange"; // gcc foolishly thinks this is reachable
}

void taskman_shutdown()
{
  taskman_flags.stop = true;
  
  for(int i = 0; i < g_num_threads; i++){
    pthread_join( taskman_flags.threads[i],NULL);
  }

  delete taskman_flags.threads;
  taskman_flags.threads = NULL;

  pthread_mutex_destroy( &barrier_lock );
}

void taskman_prepare_timers(){

  if(taskman_flags.timers_initialized){ return; }

#if defined(__linux) || defined(__linux__) || defined(linux)

  FILE* cpuinfo;
  char str[100];
  cpuinfo = fopen("/proc/cpuinfo","r");
  while(fgets(str,100,cpuinfo) != NULL){
    char cmp_str[8];
    strncpy(cmp_str, str, 7);
    cmp_str[7] = '\0';
    if(strcmp(cmp_str, "cpu MHz") == 0){
      double cpu_mhz;
      sscanf(str, "cpu MHz : %lf", &cpu_mhz);
      g_cpu_mhz = cpu_mhz;
      break;
    }
  }

  fclose( cpuinfo );
  //std::cout << "cpu MHz: " << g_cpu_mhz << std::endl;

#endif
  
  taskman_flags.timers_initialized = true;

}

void taskman_init(int num_threads)
{

  taskman_prepare_timers();

  taskman_flags.initialized = true;

  //std::cout << "taskman_init called, support thread count " << num_threads << ", queue type " ;
  //std::cout << get_queue_type_name();
  //std::cout << std::endl;

  g_num_threads = num_threads;


  if( g_queue_type == COARSE_LOCK_QUEUE ){
    CoarseLockQueue<AbstractResult*>::initStatics();
  }

  pthread_mutex_init( &barrier_lock, NULL );

#ifdef TLS_SUPPORTED
  g_task_queue = new_task_queue();
#else
  initialize_queue_key();
  pthread_setspecific( queue_key, new_task_queue() );
#endif

  taskman_flags.threads = new pthread_t[num_threads];

  for(int i=0;i<num_threads;i++){
    pthread_create(&taskman_flags.threads[i], NULL, &task_loop, reinterpret_cast<void*> (i) );
  }

#if defined(__sparc) || defined(__sparc__) || defined(sparc)
  if ( processor_bind( P_LWPID, P_MYID, chianti_binds[0], NULL ) ){
    std::cerr << "Error from processor_bind.\n" << std::endl;
  }
#endif

  bool local_sense = false;
  task_barrier( &local_sense );

  // no join
}




