#ifndef TASKMAN_H
#define TASKMAN_H

#include <boost/type_traits.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
//#include <boost/utility/result_of.hpp>

#include "AbstractQueue.h"
#include "Result.h"

// queue setup
typedef enum queue_types{

  COARSE_LOCK_QUEUE = 0,
  XACTION_QUEUE     = 1,
  FINE_LOCK_QUEUE   = 2,
  LOCK_FREE_QUEUE   = 3,
  HARDWARE_QUEUE    = 4

} queue_type_t;

const queue_type_t DEFAULT_QUEUE_TYPE = COARSE_LOCK_QUEUE;

#if defined(__linux) || defined(__linux__) || defined(linux)
extern double g_cpu_mhz;
#endif

extern const char* get_queue_type_name();

#ifdef TLS_SUPPORTED
extern __thread AbstractQueue< AbstractResult* > * g_task_queue;
#else
extern AbstractQueue< AbstractResult* >* get_g_task_queue();
#define g_task_queue get_g_task_queue()
#endif

// set_queue_type following must be called BEFORE taskman_init, if it is called at all
extern void set_queue_type(queue_type_t type);

// task initializer
extern void taskman_prepare_timers();
extern void taskman_init(int num_threads);
extern void taskman_shutdown();
extern void taskman_sync();
//extern void task_barrier(int* local_sense);

template < typename T > extern result<T>* new_result();

template < typename T > extern real_result<T>* new_real_result();

/**
 * function_ret<T>::type is the return type of T when T is a function pointer type.
 * (if T is not a function pointer type, expect goofy errors)
 * Specializations are provided for class member functions, which are not supported
 * by boost::function_traits
 */
template <typename T> struct function_ret {
    typedef typename boost::function_traits< typename boost::remove_pointer<T>::type >::result_type type;
};

template <typename T, typename Class > 
struct function_ret< T (Class::*)() >
{
    typedef T type;
};

template <typename T, typename Class, typename A1 > 
struct function_ret< T (Class::*)(A1) >
{
    typedef T type;
};

template <typename T, typename Class, typename A1, typename A2 >
struct function_ret< T (Class::*)(A1, A2) >
{
    typedef T type;
};

template <typename T, typename Class, typename A1, typename A2, typename A3 > 
struct function_ret< T (Class::*)(A1, A2, A3) >
{
    typedef T type;
};

template <typename T, typename Class, typename A1, typename A2, typename A3, typename A4 >
struct function_ret< T (Class::*)(A1, A2, A3, A4) >
{
    typedef T type;
};

/**
 * Zero-argument task function
 */
template <typename T >
result< typename boost::function_traits<T>::result_type >
task( T* func ){
    
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    //    real_result< ret_t >* real_res = new_real_result();
    real_res->future = boost::bind(func);
    real_res->addReference();

    result< ret_t >* res = new result<ret_t>(real_res);    
    g_task_queue->enqueue(res);

    return *res;
}

/**
 * One-argument task function
 */
template <typename T, typename A1>
result< typename boost::function_traits<T>::result_type >
task( T* func, A1 a1 ){
    
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    //    real_result< ret_t >* real_res = new_real_result();
    real_res->future = boost::bind(func,a1);
    real_res->addReference();

    result< ret_t >* res = new result<ret_t>(real_res);    
    g_task_queue->enqueue(res);

    return *res;
}


/**
 * Two-argument task function
 */
template <typename T, typename A1, typename A2>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2 ){
    
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;    
    real_res->future = boost::bind(func,a1,a2);
    real_res->addReference();

    result< ret_t >* res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);

    return *res;
}

/**
 * Three-argument task function
 */
template <typename T, typename A1, typename A2, typename A3>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3 ){
    
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3);
    real_res->addReference();
    
    result< ret_t >* res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);

    return *res;
}


/**
 * Four-argument task function
 */
template <typename T, typename A1, typename A2, typename A3, typename A4>
result< typename function_ret< T >::type >
task( T func, A1 a1, A2 a2, A3 a3, A4 a4 ){
    
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;    
    real_res->future = boost::bind(func,a1,a2,a3,a4);
    real_res->addReference(); 

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);

    return *res;
}

/**
 * Five-argument task function
 */ 
template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5 ){
   
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3,a4,a5);
    real_res->addReference();

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);
    
    return *res; 
}

/**
 * Six-argument task function
 */ 
template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6 ){
   
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3,a4,a5, a6);
    real_res->addReference();

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);
    
    return *res; 
}

/**
 * Seven-argument task function
 */ 
template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7 ){
   
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3,a4,a5, a6, a7);
    real_res->addReference();

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);
    
    return *res; 
}

/**
 * Eight-argument task function
 */ 
template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7, typename A8>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7, A8 a8 ){
   
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3,a4,a5, a6, a7, a8);
    real_res->addReference();

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);
    
    return *res; 
}

/**
 * Nine-argument task function
 */ 
template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7, typename A8, typename A9>
result< typename function_ret< T >::type >
task( T* func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7, A8 a8, A9 a9 ){
   
    typedef typename function_ret< T >::type ret_t;

    real_result< ret_t >* real_res = new real_result< ret_t >;
    real_res->future = boost::bind(func,a1,a2,a3,a4,a5, a6, a7, a8, a9);
    real_res->addReference();

    result< ret_t > * res = new result<ret_t>(real_res);
    g_task_queue->enqueue(res);
    
    return *res; 
}

#endif /* TASKMAN_H */
