
#include <time.h>
#include <math.h>
#include <assert.h>

/** abstract class for continuous distributions */
class Distribution
{
public:
  virtual ~Distribution(){}
  virtual const char* getName() = 0;
  virtual float getSample() = 0;
  virtual float getMean() = 0;
  virtual float getStdDeviation() = 0;

protected:
  float mGetStdUniformSample() {
    return static_cast<float> (rand()) / static_cast<float> (RAND_MAX);
  }
};

class UniformDistribution : public Distribution
{
public:
  UniformDistribution(int max) : m_max(max) { }
  ~UniformDistribution() {}

  virtual const char* getName(){
    return "Uniform";
  }
  virtual float getSample(){
    return mGetStdUniformSample() * m_max;
  }
  virtual float getMean(){
    return m_max / 2.0;
  }
  virtual float getStdDeviation(){
    return sqrt(pow(m_max,2)/12.0);
  }

private:
  int m_max;
};

class ExponentialDistribution : public Distribution
{
public:
  ExponentialDistribution(float lambda) : m_lambda(lambda) { assert(lambda > 0.0); }
  ~ExponentialDistribution(){}

  const char* getName(){
    return "Exponential";
  }
  float getSample(){
    return -1.0 * log( mGetStdUniformSample() ) / m_lambda;
  }
  float getMean(){
    return 1.0/m_lambda;
  }
  float getStdDeviation(){
    return sqrt(1/pow(m_lambda,2));
  }
  
private:
  float m_lambda;
};

class ParetoDistribution : public Distribution
{
public:
  ParetoDistribution(float alpha, float k) : m_alpha(alpha), m_k(k) {assert(alpha > 0 && k > 0);}
  ~ParetoDistribution(){}

  const char* getName(){
    return "Pareto";
  }
  float getSample(){
    return m_k / pow( mGetStdUniformSample(), (1 - m_alpha));
  }
  float getMean(){
    return (m_alpha * m_k)/(m_alpha -1);
  }
  float getStdDeviation(){
    assert(m_k > 2.0);
    return sqrt((pow(m_alpha,2)*m_k)/(pow(m_k-1,2)*(m_k-2)));
  }
private:
  float m_alpha;
  float m_k;
};

