/*
 *  Parallel LU Decomposition with Partial Pivoting
 *  lup.cilk
 *  Svetoslav Tzvetkov   <svetkins@supertech.lcs.mit.edu>
 *  6/1/98
 */

//#include <cilk-lib.cilkh>
#include <TaskMan.h>
//#include <tbb/tick_count.h>
#include "HRTimer.hh"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "getoptions.h"
#include <assert.h>

typedef double Real;

#define BLOCK_SIZE 16
/* Must be a power of 2. */

/* We are using a block row major memory. */
typedef Real Block[BLOCK_SIZE][BLOCK_SIZE];
typedef Block *Matrix;

int *P;
Matrix A;
int N;
int K;				/* size in blocks = N/BLOCK_SIZE */

#define BL(i,j) (A[(i)*K+(j)])
#define A(i,j) ((BL((i)/BLOCK_SIZE, (j)/BLOCK_SIZE))[(i)%BLOCK_SIZE][(j)%BLOCK_SIZE])

/* find the maximal elelement (by absolute value) in the column
   from rows r1..r2 */
int column_max(int c, int r1, int r2)
{
  result<int> res1, res2;
     int j = 0, k;
     int r;
     Real max = -1.0;
     Real *R;

     if (r2 - r1 < BLOCK_SIZE) {	/* can relax spawning here */
	  R = &A(r1, c);
	  for (r = r1; r <= r2;) {
	       if (fabs(*R) >= max) {
		    max = fabs(*R);
		    j = r;
	       }
	       r++;
	       R += BLOCK_SIZE;
	       if (r % BLOCK_SIZE == 0)
		    R += BLOCK_SIZE * BLOCK_SIZE * (K - 1);
	  }
     } else {
	  r = (r2 + r1) / 2;
	  res1 = task( column_max, c, r1, r);
	  res2 = task( column_max, c, r + 1, r2);
          j = *res1;
          k = *res2;
          //	  sync;
	  if (fabs(A(j, c)) < fabs(A(k, c)))
	       j = k;
     }

     return j;
}

/* update the lower rows after pivoting */
int update_lower_rows(Real * Ri, int cb, int k, int t, int b)
{
  result<int> res1, res2;
  int dummy;
     int l, m, j;
     Real a;
     Real *Rj;

     if (b - t < BLOCK_SIZE) {	/* can relax spawning here */
	  for (j = t; j <= b; j++) {
	       Rj = BL(j / BLOCK_SIZE, cb)[j % BLOCK_SIZE];
	       Rj[k] /= Ri[k];
	       a = Rj[k];
	       for (l = k + 1; l < BLOCK_SIZE; l++)
		    Rj[l] -= a * Ri[l];
	  }
     } else {
	  m = (t + b) / 2;
	  res1 = task( update_lower_rows, Ri, cb, k, t, m);
	  res2 = task( update_lower_rows, Ri, cb, k, m + 1, b);
          dummy = *res1;
          dummy = *res2;
     }
     return 0;
}

/* block LUP
   factors the block (r,c)-(r+m-1,c+m-1) and updates the
   lower rows: (r+m,c)-(N-1,c+m-1)
   performs permutations on rows r..N-1 if needed */
int block_lup(int r, int c)
{
  result<int> res1;
  int dummy;
     Real *Ri, *Rj, *R;
     Real t;
     int i, j, k, l;

     assert(r % BLOCK_SIZE == 0);
     assert(c % BLOCK_SIZE == 0);

     Ri = (Real *) BL(r / BLOCK_SIZE, c / BLOCK_SIZE);
     for (k = 0; k < BLOCK_SIZE; k++) {
	  i = r + k;

	  /* find maximum element in column */
	  res1 = task( column_max, c + k, i, N - 1);
          j = *res1;

	  /* swap rows if necessary */
	  if (i != j) {
	       l = P[i];
	       P[i] = P[j];
	       P[j] = l;
	       R = (Real *) BL(r / BLOCK_SIZE, 0) + k * BLOCK_SIZE;
	       Rj =
		   (Real *) BL(j / BLOCK_SIZE,
			       0) + (j % BLOCK_SIZE) * BLOCK_SIZE;
	       for (l = 0; l < N;) {
		    t = *R;
		    *R = *Rj;
		    *Rj = t;
		    l++;
		    R++;
		    Rj++;
		    if (l % BLOCK_SIZE == 0) {
			 R += (BLOCK_SIZE - 1) * BLOCK_SIZE;
			 Rj += (BLOCK_SIZE - 1) * BLOCK_SIZE;
		    }
	       }
	  }

	  /* update lower rows */
	  res1 = task( update_lower_rows, Ri, c / BLOCK_SIZE, k, i + 1, N - 1);
	  dummy = *res1;
	  Ri += BLOCK_SIZE;
     }
     return 0;
}

/* block schur's complement
   computes A-=B*C where
   A is (ar,ac)-(ar+m-1,ac+m-1)
   B is (br,bc)-(br+m-1,bc+m-1)
   C is (cr,cc)-(cr+m-1,cc+m-1) */
int block_schur(int ar, int ac, int br, int bc, int cr, int cc)
{
     int i, j, k;
     Real a;
     Real *RA, *R, *RB, *RC;

     assert(ar % BLOCK_SIZE == 0);
     assert(ac % BLOCK_SIZE == 0);
     assert(br % BLOCK_SIZE == 0);
     assert(bc % BLOCK_SIZE == 0);
     assert(cr % BLOCK_SIZE == 0);
     assert(cc % BLOCK_SIZE == 0);

     RB = (Real *) BL(br / BLOCK_SIZE, bc / BLOCK_SIZE);
     RA = (Real *) BL(ar / BLOCK_SIZE, ac / BLOCK_SIZE);
     for (i = 0; i < BLOCK_SIZE; i++) {
	  RC = (Real *) BL(cr / BLOCK_SIZE, cc / BLOCK_SIZE);
	  for (k = 0; k < BLOCK_SIZE; k++) {
	       R = RA;
	       a = *RB;
	       for (j = 0; j < BLOCK_SIZE; j++) {
		    *R -= *RC * a;
		    R++;
		    RC++;
	       }
	       RB++;
	  }
	  RA += BLOCK_SIZE;
     }
     return 0;
}

/* compute schur's complement: A-=B*C */
int schur(int ar, int ac, int br, int bc, int cr, int cc, int n,
		int m)
{
  int dummy;
  result<int> res1, res2, res3, res4;
     if (m == BLOCK_SIZE) {
	  int s;
          result<int>* results;
          results = new result<int>[n/m];
	  for (s = 0; s < n; s += m)
	       results[s/m] = task( block_schur, ar + s, ac, br + s, bc, cr, cc);
	  for (s = 0; s < n; s += m)
            dummy = *results[s/m];
     } else {
	  m /= 2;
	  n /= 2;

	  res1 = task( schur, ar, ac, br, bc, cr, cc, n, m);
	  res2 = task( schur, ar, ac + m, br, bc + m, cr + m, cc + m, n, m);
	  res3 = task( schur, ar + n, ac, br + n, bc + m, cr + m, cc, n, m);
	  res4 = task( schur, ar + n, ac + m, br + n, bc, cr, cc + m, n, m);
	  dummy = *res1;
	  dummy = *res2;
	  dummy = *res3;
	  dummy = *res4;

	  res1 = task( schur, ar, ac, br, bc + m, cr + m, cc, n, m);
	  res2 = task( schur, ar, ac + m, br, bc, cr, cc + m, n, m);
	  res3 = task( schur, ar + n, ac, br + n, bc, cr, cc, n, m);
	  res4 = task( schur, ar + n, ac + m, br + n, bc + m, cr + m, cc + m, n, m);
	  dummy = *res1;
	  dummy = *res2;
	  dummy = *res3;
	  dummy = *res4;
     }
     return 0;
}

/* solve: L*B=A (for B) */
int lower_solve(int lr, int lc, int ar, int ac, int m);

void block_lower_solve(int lr, int lc, int ar, int ac)
{
     int i, j, k;
     Real a;
     Real *RL, *Ri, *R, *Rk;

     assert(lr % BLOCK_SIZE == 0);
     assert(lc % BLOCK_SIZE == 0);
     assert(ar % BLOCK_SIZE == 0);
     assert(ac % BLOCK_SIZE == 0);

     RL = (Real *) BL(lr / BLOCK_SIZE, lc / BLOCK_SIZE);
     Ri = (Real *) BL(ar / BLOCK_SIZE, ac / BLOCK_SIZE);
     for (i = 1; i < BLOCK_SIZE; i++) {
	  RL += BLOCK_SIZE;
	  Ri += BLOCK_SIZE;
	  Rk = (Real *) BL(ar / BLOCK_SIZE, ac / BLOCK_SIZE);
	  for (k = 0; k < i; k++) {
	       R = Ri;
	       a = RL[k];
	       for (j = 0; j < BLOCK_SIZE; j++) {
		    *R -= *Rk * a;
		    R++;
		    Rk++;
	       }
	  }
     }
}

/* solve [L00  0 ] * [X0] = [A0]
         [L10 L11]   [X1]   [A1]  */
int lower_solve_aux(int lr, int lc, int ar, int ac, int m)
{
  result<int> res;
  int dummy;
     
  res = task( lower_solve, lr, lc, ar, ac, m);
  dummy = *res;

  res = task( schur, ar + m, ac, lr + m, lc, ar, ac, m, m);
  dummy = *res;

  res = task( lower_solve, lr + m, lc + m, ar + m, ac, m);
  dummy = *res;

     return 0;
}

/* solve [L00  0 ] * [X00 X01] = [A00 A01]
         [L10 L11]   [X10 X11]   [A10 A11] */
int lower_solve(int lr, int lc, int ar, int ac, int m)
{
  result<int> res1, res2;
  int dummy;
     if (m == BLOCK_SIZE) {
	  block_lower_solve(lr, lc, ar, ac);
     } else {
	  m /= 2;
	  res1 = task( lower_solve_aux, lr, lc, ar, ac, m);
	  res2 =  task( lower_solve_aux, lr, lc, ar, ac + m, m);
          dummy = *res1;
          dummy = *res2;
     }
     return 0;
}

/* solve a LUP subproblem of size nxm */
int lup(int r, int c, int n, int m)
{
  result<int> res1, res2, res3, res4;
  int dummy;

     if (m == BLOCK_SIZE) {
	  res1 = task( block_lup, r, c);
          dummy = *res1;
     } else {

	  m /= 2;

	  res1 = task( lup, r, c, n, m);
	  dummy = *res1;

	  res2 = task( lower_solve, r, c, r, c + m, m);
	  dummy = *res2;

	  n -= m;

	  res3 = task( schur, r + m, c + m, r + m, c, r, c + m, n, m);
	  dummy = *res3;

	  res4 = task( lup, r + m, c + m, n, m);
          dummy = *res4;
     }
     return 0;
}

/* the entry point: factors A */
int LUP(void)
{
  result<int> res;
     int i;
     for (i = 0; i < N; i++)
	  P[i] = i;
     res = task( lup, 0, 0, N, N);
     return *res;
}

/* allocate and initialize memory, etc. */
void init(int n)
{
     int i;

     /* complement n to a power of 2 */
     for (N = BLOCK_SIZE; N < n; N *= 2);
     K = N / BLOCK_SIZE;

     P = (int*) malloc(sizeof(int) * N);
     A = (Matrix) malloc(sizeof(Real) * N * N);
     assert(P != NULL);
     assert(A != NULL);

     /* initialize the permutation vector */
     for (i = 0; i < N; i++)
	  P[i] = i;
}

/* generate a random matrix */
void randomize_A(void)
{
     int i, j;
     for (i = 0; i < N; i++)
	  for (j = 0; j < N; j++)
	       A(i, j) = ((double)rand()) / (double)RAND_MAX;
}

/* read a matrix from file. the format is:
   n
   A(0,0)   ... A(0,n-1)
   .
   .
   A(n-1,0) ... A(n-1,n-1) */
void read_matrix(char *filename)
{
     FILE *fp;
     int i, j;
     int n;

     fp = fopen(filename, "rt");
     assert(fp != NULL);

     fscanf(fp, "%d", &n);
     init(n);

     for (i = 0; i < n; i++) {
	  for (j = 0; j < n; j++)
	       fscanf(fp, "%lf", &A(i, j));
	  for (; j < N; j++)
	       A(i, j) = 0.0;
     }
     for (; i < N; i++) {
	  for (j = 0; j < i; j++)
	       A(i, j) = 0.0;
	  A(i, i) = 1.0;
	  for (j = i + 1; j < N; j++)
	       A(i, j) = 0.0;
     }

     fclose(fp);
}

int usage(void)
{
     printf("\nUsage: plu <options>\n\n");
     printf("Options:\n");
     printf
	 ("  -n N : Decompose NxN matrix, where N is at least 16 and power of 2.\n");

     printf("Default: lup -n %d\n\n", 1024);

     return 1;
}

char *specifiers[] = { "-n", "-benchmark", "-h", "-nproc", 0 };
int opt_types[] = { INTARG, BENCHMARK, BOOLARG, INTARG, 0 };

/* main */
int main(int argc, char **argv)
{
     int n, benchmark, help;
     HRTimer timer;
     hrtime_t start_time, end_time;
     int nproc;
     //     Cilk_time tm_begin, tm_elapsed;
     //     Cilk_time wk_begin, wk_elapsed;
     //     Cilk_time cp_begin, cp_elapsed;

     n = 1024;
     nproc = 2;

     /* Parse arguments. */
     get_options(argc, argv, specifiers, opt_types, &n, &benchmark, &help, &nproc);

     if (help)
	  return usage();

     if (benchmark) {
	  switch (benchmark) {
	      case 1:		/* short benchmark options -- a little work */
	       n = 16;
	       break;
	      case 2:		/* standard benchmark options */
	       n = 2048;
	       break;
	      case 3:		/* long benchmark options -- a lot of work */
	       n = 2048;
	       break;
	  }
     }

     init(n);
     randomize_A();

     taskman_init(nproc-1);

     //     sync;
     //     cp_begin = Cilk_user_critical_path;
     //     wk_begin = Cilk_user_work;
     //     tm_begin = Cilk_get_wall_time();

     start_time = timer.get_time_ns();
     
     result<int> res;
     int dummy;
     res = task( LUP );
     dummy = *res;

     end_time = timer.get_time_ns();

     /* Timing. "Stop" timers */
     //     tm_elapsed = Cilk_get_wall_time() - tm_begin;
     //     wk_elapsed = Cilk_user_work - wk_begin;
     //     cp_elapsed = Cilk_user_critical_path - cp_begin;

     printf("\nCilk Example: plu\n");
     //     printf("              running on %d processor%s\n\n",
     //	    Cilk_active_size, Cilk_active_size > 1 ? "s" : "");
     printf("Options: (n x n matrix) n = %d\n\n", n);
     printf("Running time  = %lli ns\n", (end_time - start_time));
     //     printf("Work          = %4f s\n", Cilk_time_to_sec(wk_elapsed));
     //     printf("Critical path = %4f s\n\n", Cilk_time_to_sec(cp_elapsed));

     /* Free matrix. */
     free(P);
     free(A);

     taskman_shutdown();

     return 0;
}
