
#include <iostream>
#include <unistd.h>
#include <ctime>
#include <stdio.h>

#include "Distributions.h"
#include "Timers.h"
#include "TaskMan.h"

#include <tbb/task_scheduler_init.h>
#include <tbb/blocked_range.h>
#include <tbb/atomic.h>
#include <tbb/task.h>
#include <tbb/parallel_for.h>

/* test types */
#define LOOP 1
#define TAIL_TASK 2
#define HEAD_TASK 3

//using namespace std;
using namespace tbb;

atomic< hrtime_t > total_work_time;
atomic< int > total_tasks;
int g_max_depth;
int g_node_breadth;

class LoopBody 
{
  Distribution* m_dist;
public:
  LoopBody(Distribution* dist) 
    : m_dist(dist) {}
  
  void operator()( const blocked_range<int>& r) const {
    
    float work_time_fp;
    hrtime_t work_time_delta;
    hrtime_t running_time;
    hrtime_t start_time, end_time;
    
    start_time = gethrtime();
    
    work_time_fp = m_dist->getSample();
    work_time_delta = get_hrtime_delta(work_time_fp);
    
    do {
      end_time = gethrtime();
      running_time =  end_time - start_time;
    } while ( running_time < work_time_delta );
    
    end_time = gethrtime();
    
    total_work_time.fetch_and_add(end_time - start_time);
    total_tasks.fetch_and_add(1);    
  }
};
  
class HeadRecursionTask : public tbb::task 
{
  Distribution* m_dist;
  int m_depth;
public:
  HeadRecursionTask(Distribution* dist, int depth)
    : m_dist(dist), m_depth(depth) {}
  
  tbb::task* execute() {

    float work_time_fp;
    hrtime_t work_time_delta;
    hrtime_t running_time;
    hrtime_t start_time, end_time;

    if(m_depth < g_max_depth){
      set_ref_count(g_node_breadth+1);
      for(int i=0;i<g_node_breadth;i++){
	HeadRecursionTask& t = *new( allocate_child() ) HeadRecursionTask(m_dist, m_depth+1);
	if(i == g_node_breadth-1)
	  spawn_and_wait_for_all( t );
	else
	  spawn( t );
      }
    }
    
    start_time = gethrtime();
    
    work_time_fp = m_dist->getSample();
    work_time_delta = get_hrtime_delta(work_time_fp);
    
    do {
      end_time = gethrtime();
      running_time =  end_time - start_time;
    } while ( running_time < work_time_delta );
    
    end_time = gethrtime();
    
    total_work_time.fetch_and_add(end_time - start_time);
    total_tasks.fetch_and_add(1);  
    
    return NULL;
  }
};

class TailRecursionTask : public tbb::task 
{
  Distribution* m_dist;
  int m_depth;
public:
  TailRecursionTask(Distribution* dist, int depth)
    : m_dist(dist), m_depth(depth) {}
  
  tbb::task* execute() {

    float work_time_fp;
    hrtime_t work_time_delta;
    hrtime_t running_time;
    hrtime_t start_time, end_time;

    start_time = gethrtime();
    
    work_time_fp = m_dist->getSample();
    work_time_delta = get_hrtime_delta(work_time_fp);
    
    do {
      end_time = gethrtime();
      running_time =  end_time - start_time;
    } while ( running_time < work_time_delta );
    
    end_time = gethrtime();
    
    total_work_time.fetch_and_add(end_time - start_time);
    total_tasks.fetch_and_add(1);  

    if(m_depth < g_max_depth){
      set_ref_count(g_node_breadth+1);
      for(int i=0;i<g_node_breadth;i++){
	TailRecursionTask& t = *new( allocate_child() ) TailRecursionTask(m_dist, m_depth+1);
	if(i == g_node_breadth-1)
	  spawn_and_wait_for_all( t );
	else
	  spawn( t );
      }
    }
    

    
    return NULL;
  }
};

void printHelp(char* bin_name)
{
  std::cerr << "Usage: " << bin_name << " [OPTIONS]" << std::endl;
  std::cerr << "  Required: " << std::endl;
  std::cerr << "       -t <Test type> <param 1 [ param 2 ] > " << std::endl;
  std::cerr << "           Loop:         l <# tasks> " << std::endl;
  std::cerr << "           TaskTail:     t <tree depth> <node breadth> " << std::endl;
  std::cerr << "           TaskHead:     h <tree depth> <node breadth> " << std::endl;
  std::cerr << "       -d <Distribution> <param 1 [ param 2 ] > " << std::endl;
  std::cerr << "           Uniform:      u <max>" << std::endl;
  std::cerr << "           Exponential:  e <lambda>" << std::endl;
  std::cerr << "           Pareto:       p <alpha> <k>" << std::endl;
  std::cerr << "       -n <# threads> " << std::endl << std::endl;
  std::cerr << "  Optional:" << std::endl;
  std::cerr << "       -s <seed value | 0 for random>" << std::endl;
  std::cerr << "       -q <queue type>" << std::endl;
  std::cerr << "           FineGrainedLock: l {default} " << std::endl;
  std::cerr << "           Xaction:         x " << std::endl;
  std::cerr << "           Lock-Free:       w " << std::endl;
}

int main(int argc, char* argv[])
{
  int opt;
  Distribution* dist = NULL;
  int num_tasks, num_threads;
  int test_type = LOOP;
  int seed;

  total_work_time = 0;
  total_tasks = 0;

  num_tasks = 10000; // default to something
  num_threads = 3;
  seed = time(NULL);

  set_queue_type(COARSE_LOCK_QUEUE);

  if( argc < 2 ){
    printHelp(argv[0]);
    exit(EXIT_FAILURE);
  }
  while((opt = getopt(argc, argv, "d:n:t:s:q:h")) != -1){
    switch(opt){
    case 'd':
      if(optarg[0] == 'e'){
        float lambda;
        lambda = atof( argv[optind] );
        optind++;
        dist = new ExponentialDistribution(lambda);
      } else if(optarg[0] == 'p'){
        float alpha, k;
        alpha = atof( argv[optind] );
        assert(optind <= argc-1);
        optind++;
        k = atof( argv[optind] );
        optind++;
        dist = new ParetoDistribution(alpha, k);
      } else if(optarg[0] == 'u'){
        int max;
        max = atoi( argv[optind] );
        optind++;
        dist = new UniformDistribution(max);
      } else {
        std::cerr << "Unkown distribution type:" << optarg << std::endl;
        exit(EXIT_FAILURE);
      }
      break;
    case 'h':
      printHelp(argv[0]);
      exit(EXIT_SUCCESS);
    case 'n':
      num_threads = atoi(optarg);
      break;
    case 's':
      seed = atoi(optarg);
      break;
    case 't':
      if(optarg[0] == 'l'){
        test_type = LOOP;
        num_tasks = atoi( argv[optind] );
        optind++;
      } else if(optarg[0] == 't'){
        test_type = TAIL_TASK;
        g_max_depth = atoi( argv[optind] );
        optind++;
        g_node_breadth = atoi( argv[optind] );
        optind++;
      } else if(optarg[0] == 'h'){
        test_type = HEAD_TASK;
        g_max_depth = atoi( argv[optind] );
        optind++;
        g_node_breadth = atoi( argv[optind] );
        optind++;
      } else {
        std::cerr << "Unkown test type:" << optarg << std::endl;
        exit(EXIT_FAILURE);
      }
      break;
    case 'q':
      if( optarg[0] == 'l'){
        set_queue_type(COARSE_LOCK_QUEUE);
        optind++;
        break;
      } else if( optarg[0] == 'x' ){
        set_queue_type(XACTION_QUEUE);
        optind++;
      } else if( optarg[0] == 'w' ){
        set_queue_type(LOCK_FREE_QUEUE);
        optind++;
      } else {
        std::cerr << "Unkown queue type: " << optarg << std::endl;
        exit(EXIT_FAILURE);
      }
      break;
    default:
      std::cerr << "Unkown Argument: " << opt << "\n";
      exit(EXIT_FAILURE);
    };
  }
  if( dist == NULL ){ printHelp(argv[0]); }

  srand(seed);

  std::cout << "Running with parameters: " << std::endl;
  std::cout << "   Test: " << test_type << std::endl;
  std::cout << "   " << dist->getName() << " distribution - mean: " << dist->getMean() << " std deviation: " << dist->getStdDeviation() << std::endl;
  std::cout << "   Queue Type: " << get_queue_type_name() << std::endl;
  std::cout << "   Seed value: " << seed << std::endl;
  std::cout << "   Number of tasks:   " << num_tasks << std::endl;
  std::cout << "   Number of threads: " << num_threads << std::endl;

  task_scheduler_init init( num_threads );
  taskman_prepare_timers();


  //int total;

  hrtime_t start_time, end_time;

  start_time = gethrtime();

  if(test_type == LOOP){
    LoopBody loop(dist);
    tbb::parallel_for(tbb::blocked_range<int>(0,num_tasks, 1), loop );
  } else if(test_type == TAIL_TASK){
    TailRecursionTask& t = *new(tbb::task::allocate_root()) TailRecursionTask(dist, 0);
    tbb::task::spawn_root_and_wait(t);
  } else if(test_type == HEAD_TASK){
    HeadRecursionTask& t = *new(tbb::task::allocate_root()) HeadRecursionTask(dist, 0);
    tbb::task::spawn_root_and_wait(t);
  } else {
    std::cerr << "Unkown test type" << std::endl;
    exit(EXIT_FAILURE);
  }

  end_time = gethrtime();
  std::cout << "Total runtime (s):       " << get_time_delta_s(end_time, start_time) << std::endl;
  std::cout << "Total working time (s):  " << get_time_delta_s(total_work_time, 0) << std::endl;
  std::cout << "Total tasks executed:    " << total_tasks << std::endl;

  return EXIT_SUCCESS;

}
